/*
 * ping.c - send ICMP Echo request packets
 *
 * Copyright (c) 2000 Tim J. Robbins. 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: ping.c,v 1.3 2000/11/18 22:31:20 fyre Stab $
 */

#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>

static char *rcsid = "$Id: ping.c,v 1.3 2000/11/18 22:31:20 fyre Stab $";

static int in_cksum (unsigned short *buf, int len);
static void make_addr (const char *host, struct sockaddr_in *sa);
static void send_echo (struct sockaddr_in *dest);
static void usage (void);
static int wait_response (struct sockaddr_in *src);

static int rawfd;
static int timeout = 20;

int main (int argc, char *argv[])
{
	struct sockaddr_in to;
	const char *host;
	int fd;

	/*
	 * Temporarily drop our suid root privileges until we need them.
	 */
	if (seteuid (getuid ()) < 0) {
		fprintf (stderr, "ping: seteuid: %s\n", strerror (errno));
		exit (EXIT_FAILURE);
	}

	/*
	 * Ensure that the raw socket we're about to open will not be
	 * file descriptor 0, 1 or 2 for security reasons.
	 */
	do {
		if ((fd = open ("/dev/null", O_RDWR)) < 0) {
			fprintf (stderr, "ping: /dev/null: %s\n",
					strerror (errno));
			exit (EXIT_FAILURE);
		}
	} while (fd <= 2);
	if (close (fd) < 0) {
		fprintf (stderr, "ping: close: %s\n", strerror (errno));
		exit (EXIT_FAILURE);
	}

	/*
	 * ping needs root privileges to open the raw ICMP socket. Once the
	 * socket is open, we permanently drop our privileges.
	 */
	if (seteuid (0) < 0) {
		fprintf (stderr, "ping: seteuid 0: %s\n", strerror (errno));
		exit (EXIT_FAILURE);
	}
	
	if ((rawfd = socket (PF_INET, SOCK_RAW, IPPROTO_ICMP)) < 0) {
		fprintf (stderr, "ping: raw socket: %s\n", strerror (errno));
		exit (EXIT_FAILURE);
	}
	if (rawfd <= 2) exit (EXIT_FAILURE); /* should not happen */

	if (setuid (getuid ()) < 0) {
		fprintf (stderr, "ping: setuid: %s\n", strerror (errno));
		exit (EXIT_FAILURE);
	}

	/*
	 * We are now running as the calling user and have opened the raw
	 * ICMP socket..
	 */

	if ((argc != 2) && (argc != 3))
		usage ();
	host = argv[1];
	if (argc > 2) {
		timeout = atoi (argv[2]);
		if (timeout <= 0) {
			fprintf (stderr, "ping: illegal timeout value\n");
			exit (EXIT_FAILURE);
		} else if (timeout > 60 * 60) {
			fprintf (stderr, "ping: illegal timeout value\n");
			exit (EXIT_FAILURE);
		}
	}

	make_addr (host, &to);

	send_echo (&to);
	if (wait_response (&to) != 0) {
		printf ("%s did not respond\n", host);
		exit (EXIT_FAILURE);
	}

	printf ("%s is alive\n", host);

	return EXIT_SUCCESS;
}

/*
 * Send an ICMP_ECHO request to `dest'.
 */
static void send_echo (struct sockaddr_in *dest)
{
	unsigned char outpacket[sizeof (struct icmphdr)];
	struct icmphdr *ich;
	int nw;

	ich = (struct icmphdr *) outpacket;
	ich->type = ICMP_ECHO;
	ich->code = 0;
	ich->checksum = 0;
	ich->un.echo.sequence = 0;
	ich->un.echo.id = getpid() & 0xFFFF;
	ich->checksum = in_cksum ((u_short *) outpacket, sizeof (outpacket));

	if ((nw=sendto (rawfd, outpacket, sizeof (outpacket), 0, dest, sizeof (*dest))) < 0) {
		fprintf (stderr, "ping: sendto: %s\n", strerror (errno));
		exit (EXIT_FAILURE);
	}

	if (nw != sizeof (outpacket)) {
		fprintf (stderr, "ping: short write\n");
		exit (EXIT_FAILURE);
	}
}

/*
 * Convert a host string (hostname or IP address) to a sockaddr_in structure.
 */
static void make_addr (const char *host, struct sockaddr_in *sa)
{
	struct hostent *he;
	
	sa->sin_port = 0;
	sa->sin_family = AF_INET;

	if ((he = gethostbyname (host)) != NULL) {
		if (he->h_addrtype != AF_INET) {
			fprintf (stderr, "ping: host is not an INET address\n");
			exit (EXIT_FAILURE);
		}
		if (he->h_addr_list[0] == NULL) {
			fprintf (stderr, "ping: host has no addresses\n");
			exit (EXIT_FAILURE);
		}
		memcpy (&sa->sin_addr, he->h_addr_list[0], sizeof (sa->sin_addr));
		if (he->h_addr_list[1] != NULL) {
			fprintf (stderr, "ping: warning: host has multiple addresses, using %s\n",
				inet_ntoa (sa->sin_addr));
		}
	} else if (inet_aton (host, &sa->sin_addr) == 0) {
		fprintf (stderr, "ping: unknown host `%s'\n", host);
		exit (EXIT_FAILURE);
	}
}

void ms_to_tv (int ms, struct timeval *tv)
{
	tv->tv_sec = ms / 1000;
	tv->tv_usec = (ms % 1000) * 1000;
}

int tv_to_ms (const struct timeval *tv)
{
	return tv->tv_sec * 1000 + tv->tv_usec / 1000;
}

/*
 * Wait for an ICMP_ECHOREPLY from the host.
 */
static int wait_response (struct sockaddr_in *src)
{
	struct sockaddr_in from;
	socklen_t fromlen = sizeof (from);
	struct iphdr *iph;
	struct icmphdr *ich;
	unsigned char inpacket[IP_MAXPACKET];
	fd_set fdset;
	int toremain = timeout * 1000;
	struct timeval totval, before, after;

	/*
	 * Keep reading packets until we find the response to the packet
	 * we sent or timeout.
	 */
	do {
		FD_ZERO (&fdset);
		FD_SET (rawfd, &fdset);
		
		if (gettimeofday (&before, NULL) < 0) {
			fprintf (stderr, "ping: gettimeofday: %s\n",
					strerror (errno));
			exit (EXIT_FAILURE);
		}
		ms_to_tv (toremain, &totval);
		if (select (rawfd + 1, &fdset, NULL, NULL, &totval) < 0) {
			fprintf (stderr, "ping: select: %s\n", strerror (errno));
			exit (EXIT_FAILURE);
		}
		if (gettimeofday (&after, NULL) < 0) {
			fprintf (stderr, "ping: gettimeofday: %s\n",
					strerror (errno));
			exit (EXIT_FAILURE);
		}
		toremain -= tv_to_ms (&after) - tv_to_ms (&before);
		if (toremain < 0)
			return 1;
		if (!FD_ISSET (rawfd, &fdset))
			return 1;
		if (recvfrom (rawfd, inpacket, sizeof (inpacket), 0, &from, &fromlen) < 0) {
			fprintf (stderr, "ping: recvfrom: %s\n", strerror (errno));
			exit (EXIT_FAILURE);
		}
		iph = (struct iphdr *) inpacket;
		ich = (struct icmphdr *) (inpacket + iph->ihl * 4);
	} while ((from.sin_addr.s_addr != src->sin_addr.s_addr) || 
		(ich->type != ICMP_ECHOREPLY) || 
		(ich->un.echo.id != (getpid() & 0xFFFF)) || 
		(ich->un.echo.sequence != 0));
	
	return 0;
}

/*
 * Calculate IP checksum.
 */
static int in_cksum (unsigned short *buf, int len)
{
	int sum = 0;

	while (len > 1) {
		sum += *buf++;
		len -= 2;
	}
	
	if (len == 1) {
		union {
			unsigned short s;
			unsigned char c[2];
		} u;
		u.c[0] = *(unsigned char *) buf;
		u.c[1] = 0;
		sum += u.s;
	}

	sum = (sum >> 16) + (sum & 0xFFFF);
	sum += (sum >> 16);
	return ~sum;
}

static void usage (void)
{
	fprintf (stderr, "usage: ping host [timeout]\n");
	exit (EXIT_FAILURE);
}
